#ifndef PEBBLISP_PLFUNC_H
#define PEBBLISP_PLFUNC_H

#include "pebblisp.h"

fn(add, "+",
   "Add numbers",
   "(+ 1 2)", "3",
   "(+ 1 2 3 4)", "10",
);

fn(sub, "-",
   "Subtract numbers",
   "(- 3 2)", "1",
   "(- 10 3 1)", "6",
);

fn(mul, "*",
   "Multiply numbers",
   "(* 3 2)", "6",
   "(* 1 2 3 4)", "24",
);

fn(dvi, "/",
   "Divide numbers",
   "(/ 8 2)", "4",
   "(/ 100 5 2)", "10",
);

fn(mod, "%",
   "Get the modulus of two numbers",
   "(% 10 3)", "1",
);

fn(equ, "=",
   "Check if two values are equal",
   "(= (+ 2 2) 4)", "T",
   "(= \"Hello\" (cat \"He\" \"llo\"))", "T",
   "(= (1 2 3) (1 2 3))", "T",
   "(= 10 \"10\")", "F",
   "(= 2 5)", "F",
);

fn(greaterThan, ">",
   "Check if one number is greater than another",
   "(> 10 3)", "T",
   "(> 2 5)", "F",
);

fn(lessThan, "<",
   "Check if one number is less than another",
   "(< 2 5)", "T",
   "(< 10 3)", "F",
);

fn(and, "&",
   "Check if all conditions are true",
   "(& T T)", "T",
   "(& F T)", "F",
   "(& T T T T T F)", "F",
);

fn(or, "|",
   "Check if any conditions are true",
   "(| T F)", "T",
   "(| F F)", "F",
   "(| F F F F F T)", "T",
);

tfn(catObjects, "cat",
    ({ NULL, isStringy }),
    "Concatenate string representations of the given objects.",
    "(cat \"Stuff: \" (1 2 3))", "Stuff: ( 1 2 3 )",
);

tfn(filter, "fil",
    ({ isFuncy, isListy, isListy }),
    "Filter a list based on the given condition.",
    "(fil (fn (a) (< 50 a)) (25 60 100))", "( 60 100 )",
    "(fil (fn (a) (< 0 (len a))) ( () (1) (1 2) () ))", "( ( 1 ) ( 1 2 ) )",
);

tfn(append, "ap",
    ({ isListy, NULL, isListy }),
    "Append the given element. Creates a new list.",
    "(ap (1 2) 3)", "( 1 2 3 )",
);

tfn(prepend, "pre",
    ({ isListy, NULL, isListy }),
    "Prepend the given element. Creates a new list",
    "(pre (2 3) 1)", "( 1 2 3 )",
);

tfn(len, "len",
    ({ isListy, isNumber }),
    "Returns the length of the given list, or an error if the expression is not a list.",
    "(len (2 3))", "2",
    "(len ())", "0",
    "(len \"string\")", "BAD_PARAMS_ON: len",
);

tfn(reduce, "reduce",
    ({ NULL, isFuncy, NULL, NULL }),
    "Performs a simple reduction. Does not currently work with lambdas.\n"
    "Takes three arguments:\n"
    " - Values\n"
    " - A function to apply to each value\n"
    " - An initial value",
    "(reduce 5 + 6)", "11",
    "(reduce (1 2 3) + 0)", "6",
);

tfn(at, "at",
    ({ isNumber, isListy, NULL }),
    "Get item at the given index in the given list.",
    "(at 1 (1 2 3))", "2",
    "(at 99 (1 2 3))", "INDEX_PAST_END",
    "(at 99 \"string\")", "BAD_PARAMS_ON: at",
);

tfn(rest, "rest",
    ({ isListy, isListy }),
    "Get the tail of a list. All but the first element.",
    "(rest (1 2 3))", "( 2 3 )",
    "(rest ())", "( )",
    "(rest \"string\")", "BAD_PARAMS_ON: rest",
);

tfn(reverse, "rev",
    ({ isListy, isListy }),
    "Reverse a list.",
    "(rev (1 2 3))", "( 3 2 1 )",
    "(rev \"string\")", "BAD_PARAMS_ON: reverse",
);

tfn(isNum, "isnum",
    ({ NULL, isBool }),
    "Returns `T` only if the argument evaluates to a number.",
    "(isnum 1)", "T",
    "(isnum (+ 5 5))", "T",
    "(isnum '(+ 5 5))", "F",
    "(isnum \"Hello\")", "F",
);

tfn(isList, "islist",
    ({ NULL, isBool }),
    "Returns `T` only if the argument is a list.",
    "(islist (1 2 3))", "T",
    "(islist ())", "T",
    "(islist \"Stringy\")", "F",
);

tfn(isString, "isstr",
    ({ NULL, isBool }),
    "Returns `T` only if the argument is a string.",
    "(isstr \"Heyo\")", "T",
    "(isstr \"\")", "T",
    "(isstr (cat 5 5))", "T",
    "(isstr 10)", "F",
);

tfn(isErr, "iserr",
    ({ NULL, isBool }),
    "Check if the argument is an error.",
    "(iserr (at 10 ()))", "T",
    "(iserr 5)", "F",
);

tfn(charAt, "chat",
    ({ isStringy, isNumber, isStringy }),
    "Get the char in the given string at the given index.",
    "(chat \"Hello\" 1)", "e",
    "(chat \"Hello\" 10)", "",
);

tfn(charVal, "char",
    ({ isStringy, isNumber }),
    "Get the ascii integer representaton of the given character.",
    "(char \"h\")", "104",
    "(char \"hello\")", "104",
    "(char \"\")", "0",
);

/// STRING/SLIST => ANY
fn(parseEvalO, "eval",
   "Evaluate the given string or quoted list.",
   "(eval \"(1 2 3)\")", "( 1 2 3 )",
   "(eval '(+ 5 5))", "10",
);

#ifdef STANDALONE

fn(print, "prn", "Prints the string representation of the given object to stdout.");

tfn(numToChar, "ch",
    ({ isNumber, isStringy }),
    "Gets a string containing the ascii character for the given number value.",
    "(ch 107)", "k",
    "(ch 0x21)", "!",
);

fn(printEnvO, "penv",
    "Prints out the current scoped environment.\n"
    "(penv) prints a mostly human-readable list of env variables.\n"
    "(penv T) prints a mostly list of env variables, including pointer addresses.\n"
    "Calling (penv) with no argument is equivalent to (penv F)"
);

tfn(systemCall, "sys",
    ({ isStringy, isNumber }),
    "Opens a shell and runs the given command, returning 0 if successful.\n"
    "If the argument is not a string, returns 255.\n",
    "(sys \"echo yee > /dev/null\")", "0",
);

tfn(loadFile, "loadfile",
    ({ isStringy, NULL }),
    "Loads and parses the given file.\n"
    "Returns 0 if the file was loaded and parsed successfully. Otherwise 1.\n"
    "(loadfile \"printdate.pl\")\n"
    "Mon 21 Mar 2022 10:35:03 AM EDT\n"
    "=> 0"
);

/// @code
/// () => STRING
/// STRING => STRING
fn(takeInput, "inp",
   "Take console input with an optional prompt. For example:\n"
   "`(def x (input))` will wait for user input with no prompt.\n"
   "`(def x (input \">> \"))` wait for input, but prompt the user with '>> '.\n"
);

tfn(readFileToObject, "rf",
    ({ isStringy, isStringy }),
    "Read a file into a string object."
);

tfn(getEnvVar, "env",
    ({ isStringy, isStringy }),
    "Get a variable from the current environment\n"
    "(env HOME) => /home/sagevaillancourt"
);

tfn(getTime, "time",
    ({ isStruct }),
    "Get a struct of the current time with fields (minute hour sec)."
);

#endif // STANDALONE

#endif // PEBBLISP_PLFUNC_H